@@ -26,6 +26,7 @@
 #include "arch/common/fw_cfg.h"
 #include "arch/sparc64/ofmem_sparc64.h"
 #include "spitfire.h"
+#include "libc/vsprintf.h"
 
 #define UUID_FMT "%02x%02x%02x%02x-%02x%02x-%02x%02x-%02x%02x-%02x%02x%02x%02x%02x%02x"
 
@@ -60,7 +61,7 @@ static const struct hwdef hwdefs[] = {
             .cfg_base = APB_SPECIAL_BASE,
             .cfg_len = 0x1000000,
             .host_pci_base = APB_MEM_BASE,
-            .pci_mem_base = 0x100000, /* avoid VGA at 0xa0000 */
+            .pci_mem_base = 0x20000000, /* avoid VGA at 0xa0000 */
             .mem_len = 0xf0000000,
             .io_base = APB_SPECIAL_BASE + 0x2000000ULL, // PCI Bus I/O space
             .io_len = 0x1000000,
@@ -120,6 +121,28 @@ sparc64_reset_all(void)
         : : "r" (val), "r" (addr) : "memory");
 }
 
+/* Power off */
+static void
+sparc64_power_off(void)
+{
+    /* Locate address of ebus power device */
+    phandle_t ph;
+    uint32_t addr;
+    volatile uint32_t *p;
+    int len;
+
+    ph = find_dev("/pci/pci@1,1/ebus/power");
+    if (ph) {
+        addr = get_int_property(ph, "address", &len);
+
+        if (len) {
+            /* Set bit 24 to invoke power off */
+            p = cell2pointer(addr);
+            *p = 0x1000000;
+        }
+    }
+}
+
 /* PCI Target Address Space Register (see UltraSPARC IIi User's Manual
   section 19.3.0.4) */
 #define PBM_PCI_TARGET_AS              0x2028
@@ -516,6 +539,8 @@ void arch_nvram_get(char *data)
     uint32_t clock_frequency;
     uint16_t machine_id;
     const char *stdin_path, *stdout_path;
+    char *bootorder_file, *boot_path;
+    uint32_t bootorder_sz, sz;
 
     fw_cfg_init();
 
@@ -535,9 +560,15 @@ void arch_nvram_get(char *data)
     }
 
     kernel_size = fw_cfg_read_i32(FW_CFG_KERNEL_SIZE);
-    if (kernel_size)
+    if (kernel_size) {
         kernel_image = fw_cfg_read_i64(FW_CFG_KERNEL_ADDR);
 
+        /* Mark the kernel memory as mapped 1:1 and in use */
+        ofmem_claim_phys(PAGE_ALIGN(kernel_image), PAGE_ALIGN(kernel_size), 0);
+        ofmem_claim_virt(PAGE_ALIGN(kernel_image), PAGE_ALIGN(kernel_size), 0);
+        ofmem_map(PAGE_ALIGN(kernel_image), PAGE_ALIGN(kernel_image), PAGE_ALIGN(kernel_size), -1);
+    }
+
     size = fw_cfg_read_i32(FW_CFG_CMDLINE_SIZE);
     if (size) {
 	obio_cmdline = (char *)malloc(size + 1);
@@ -548,7 +579,16 @@ void arch_nvram_get(char *data)
     }
     qemu_cmdline = (uint64_t)obio_cmdline;
     cmdline_size = size;
-    boot_device = fw_cfg_read_i16(FW_CFG_BOOT_DEVICE);
+
+    initrd_size = fw_cfg_read_i32(FW_CFG_INITRD_SIZE);
+    if (initrd_size) {
+        initrd_image = fw_cfg_read_i32(FW_CFG_INITRD_ADDR);
+
+        /* Mark initrd memory as mapped 1:1 and in use */
+        ofmem_claim_phys(PAGE_ALIGN(initrd_image), PAGE_ALIGN(initrd_size), 0);
+        ofmem_claim_virt(PAGE_ALIGN(initrd_image), PAGE_ALIGN(initrd_size), 0);
+        ofmem_map(PAGE_ALIGN(initrd_image), PAGE_ALIGN(initrd_image), PAGE_ALIGN(initrd_size), -1);
+    }
 
     if (kernel_size)
         printk("kernel addr %llx size %llx\n", kernel_image, kernel_size);
@@ -608,7 +648,11 @@ void arch_nvram_get(char *data)
     push_str("/options");
     fword("find-device");
 
-    switch (boot_device) {
+    /* Boot order */
+    bootorder_file = fw_cfg_read_file("bootorder", &bootorder_sz);
+
+    if (bootorder_file == NULL) {
+        switch (fw_cfg_read_i16(FW_CFG_BOOT_DEVICE)) {
         case 'a':
             push_str("/obio/SUNW,fdtwo");
             break;
@@ -622,11 +666,31 @@ void arch_nvram_get(char *data)
         case 'n':
             push_str("net");
             break;
-    }
+        }
 
-    fword("encode-string");
-    push_str("boot-device");
-    fword("property");
+        fword("encode-string");
+        push_str("boot-device");
+        fword("property");
+    } else {
+        sz = bootorder_sz * (3 * 2);
+        boot_device = malloc(sz);
+        memset(boot_device, 0, sz);
+
+        while ((boot_path = strsep(&bootorder_file, "\n")) != NULL) {
+            snprintf(buf, sizeof(buf),
+                     "%s:f "
+                     "%s:a "
+                     "%s ",
+                     boot_path, boot_path, boot_path);
+
+            strncat(boot_device, buf, sz);
+        }
+
+        push_str(boot_device);
+        fword("encode-string");
+        push_str("boot-device");
+        fword("property");
+    }
 
     push_str(obio_cmdline);
     fword("encode-string");
@@ -700,6 +764,40 @@ static void init_memory(void)
     PUSH(virt + MEMORY_SIZE);
 }
 
+/* ( size -- virt ) */
+static void
+dma_alloc(void)
+{
+    ucell size = POP();
+    ucell addr;
+    int ret;
+
+    /* OpenBIOS doesn't enable the sun4u IOMMU so we can fall back to
+     * using ofmem_posix_memalign */
+    ret = ofmem_posix_memalign((void *)&addr, size, PAGE_SIZE);
+
+    if (ret) {
+        PUSH(0);
+    } else {
+        PUSH(addr);
+    }
+}
+
+/* ( virt devaddr size -- ) */
+static void
+dma_sync(void)
+{
+    ucell size = POP();
+    POP();
+    ucell virt = POP();
+    ucell va;
+
+    for (va = virt; va < virt + size; va += PAGE_SIZE_8K) {
+        itlb_demap(va);
+        dtlb_demap(va);
+    }
+}
+
 extern volatile uint64_t *obp_ticks_pointer;
 
 static void
@@ -707,6 +805,12 @@ arch_init( void )
 {
 	openbios_init();
 	modules_init();
+
+    bind_func("sparc64-dma-alloc", dma_alloc);
+    feval("['] sparc64-dma-alloc to (dma-alloc)");
+    bind_func("sparc64-dma-sync", dma_sync);
+    feval("['] sparc64-dma-sync to (dma-sync)");
+
 #ifdef CONFIG_DRIVER_PCI
         ob_pci_init();
 
@@ -727,6 +831,11 @@ arch_init( void )
         bind_func("spacel@", spacel_read);
         bind_func("spacex@", spacex_read);
 
+        /* Bind power functions */
+        bind_func("sparc64-power-off", sparc64_power_off);
+        push_str("' sparc64-power-off to power-off");
+        fword("eval");
+
 	bind_func("platform-boot", boot );
 }
 
